import os
import sys
import tempfile
import contextlib

import shutil
import distutils

import setuptools
import setuptools.command.build_ext
import cppimport.config
from cppimport.filepaths import make_absolute

if sys.version_info[0] == 2:
    import StringIO as io
else:
    import io


@contextlib.contextmanager
def stdchannel_redirected(stdchannel):
    """
    Redirects stdout or stderr to a StringIO object. As of python 3.4, there is a
    standard library contextmanager for this, but backwards compatibility!
    """
    try:
        s = io.StringIO()
        old = getattr(sys, stdchannel)
        setattr(sys, stdchannel, s)
        yield s
    finally:
        setattr(sys, stdchannel, old)


# Subclass setuptools Extension to add a parameter specifying where the shared
# library should be placed after being compiled
class ImportCppExt(setuptools.Extension):
    def __init__(self, libdest, *args, **kwargs):
        self.libdest = libdest
        setuptools.Extension.__init__(self, *args, **kwargs)


# Subclass setuptools build_ext to put the compiled shared library in the
# appropriate place in the source tree.
class BuildImportCppExt(setuptools.command.build_ext.build_ext):
    def copy_extensions_to_source(self):
        for ext in self.extensions:
            fullname = self.get_ext_fullname(ext.name)
            filename = self.get_ext_filename(fullname)
            src_filename = os.path.join(self.build_lib, filename)
            dest_filename = os.path.join(ext.libdest, os.path.basename(filename))

            distutils.file_util.copy_file(
                src_filename, dest_filename, verbose=self.verbose, dry_run=self.dry_run
            )


# Patch for parallel compilation with distutils
# From: http://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils # noqa: E501
def parallel_compile(
    self,
    sources,
    output_dir=None,
    macros=None,
    include_dirs=None,
    debug=0,
    extra_preargs=None,
    extra_postargs=None,
    depends=None,
):

    # these lines are copied directly from distutils.ccompiler.CCompiler
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs
    )
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

    # Determine the number of compilation threads. Unless there are special
    # circumstances, this is the number of cores on the machine
    N = 1
    try:
        import multiprocessing
        import multiprocessing.pool

        N = multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    def _single_compile(obj):
        try:
            src, ext = build[obj]
        except KeyError:
            return
        # import time
        # start = time.time()
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
        # end = time.time()
        # print("took " + str(end - start) + " to compile " + str(obj))

    # imap is evaluated on demand, converting to list() forces execution
    list(multiprocessing.pool.ThreadPool(N).imap(_single_compile, objects))
    return objects


def build_module(module_data):
    build_path = tempfile.mkdtemp()

    full_module_name = module_data["fullname"]
    filepath = module_data["filepath"]
    cfg = module_data["cfg"]

    module_data["abs_include_dirs"] = [
        make_absolute(module_data["filedirname"], d)
        for d in cfg.get("include_dirs", [])
    ] + [os.path.dirname(filepath)]
    module_data["abs_library_dirs"] = [
        make_absolute(module_data["filedirname"], d)
        for d in cfg.get("library_dirs", [])
    ]
    module_data["dependency_dirs"] = module_data["abs_include_dirs"] + [
        module_data["filedirname"]
    ]
    module_data["extra_source_filepaths"] = [
        make_absolute(module_data["filedirname"], s) for s in cfg.get("sources", [])
    ]

    ext = ImportCppExt(
        os.path.dirname(filepath),
        full_module_name,
        language="c++",
        sources=(
            module_data["extra_source_filepaths"]
            + [module_data["rendered_src_filepath"]]
        ),
        include_dirs=module_data["abs_include_dirs"],
        extra_compile_args=cfg.get("extra_compile_args", []),
        extra_link_args=cfg.get("extra_link_args", []),
        library_dirs=module_data["abs_library_dirs"],
        libraries=cfg.get("libraries", []),
    )

    args = ["build_ext", "--inplace"]
    args.append("--build-temp=" + build_path)
    args.append("--build-lib=" + build_path)

    if cppimport.config.quiet:
        args.append("-q")
    else:
        args.append("-v")

    setuptools_args = dict(
        name=full_module_name,
        ext_modules=[ext],
        script_args=args,
        cmdclass={"build_ext": BuildImportCppExt},
    )

    # Monkey patch in the parallel compiler if requested.
    py33orgreater = sys.version_info[0] >= 3 and sys.version_info[1] >= 3
    parallelize = cfg.get("parallel") and py33orgreater
    if parallelize:
        old_compile = distutils.ccompiler.CCompiler.compile
        distutils.ccompiler.CCompiler.compile = parallel_compile

    if cppimport.config.quiet:
        with stdchannel_redirected("stdout"):
            with stdchannel_redirected("stderr"):
                setuptools.setup(**setuptools_args)
    else:
        setuptools.setup(**setuptools_args)

    # Remove the parallel compiler to not corrupt the outside environment.
    if parallelize:
        distutils.ccompiler.CCompiler.compile = old_compile

    shutil.rmtree(build_path)



import ctypes

quiet = True
should_force_rebuild = False
file_exts = [".cpp", ".c"]
rtld_flags = ctypes.RTLD_LOCAL


def set_quiet(to):
    global quiet
    quiet = to


def force_rebuild(to=True):
    global should_force_rebuild
    should_force_rebuild = to


def quiet_print(a):
    global quiet
    if not quiet:
        print(a)


def turn_off_strict_prototypes():
    import distutils.sysconfig

    cfg_vars = distutils.sysconfig.get_config_vars()
    for key, value in cfg_vars.items():
        if type(value) == str:
            cfg_vars[key] = value.replace("-Wstrict-prototypes", "")


def set_rtld_flags(flags):
    global rtld_flags
    rtld_flags = flags



import sys
import traceback
import cppimport.importer


class Hook(object):
    def __init__(self):
        self._running = False
        self.print_exceptions = False

    def find_module(self, fullname, path=None):
        # Prevent re-entry by the underlying importer
        if self._running:
            return

        try:
            self._running = True
            cppimport.importer.imp(fullname, opt_in=True)
        except ImportError:
            # ImportError should be quashed because that simply means cppimport
            # didn't find anything, and probably shouldn't have found anything!
            if self.print_exceptions:
                print(traceback.format_exc())
        finally:
            self._running = False


hook_obj = Hook()
sys.meta_path.insert(0, hook_obj)



import os
import sys

import cppimport.config


def check_contains_cppimport(filepath):
    with open(filepath, "r") as f:
        return "cppimport" in f.readline()


def find_file_in_folders(filename, paths, opt_in):
    for d in paths:
        if not os.path.exists(d):
            continue

        if os.path.isfile(d):
            continue

        for f in os.listdir(d):
            if f != filename:
                continue
            filepath = os.path.join(d, f)
            if opt_in and not check_contains_cppimport(filepath):
                continue
            return filepath
    return None


def find_matching_path_dirs(moduledir):
    if moduledir == "":
        return sys.path

    ds = []
    for dir in sys.path:
        test_path = os.path.join(dir, moduledir)
        if os.path.exists(test_path) and os.path.isdir(test_path):
            ds.append(test_path)
    return ds


def _find_module_cpppath(modulename, opt_in=False):
    modulepath_without_ext = modulename.replace(".", os.sep)
    moduledir = os.path.dirname(modulepath_without_ext + ".throwaway")
    matching_dirs = find_matching_path_dirs(moduledir)
    matching_dirs = [os.getcwd() if d == "" else d for d in matching_dirs]
    matching_dirs = [
        d if os.path.isabs(d) else os.path.join(os.getcwd(), d) for d in matching_dirs
    ]

    for ext in cppimport.config.file_exts:
        modulefilename = os.path.basename(modulepath_without_ext + ext)
        outfilename = find_file_in_folders(modulefilename, matching_dirs, opt_in)
        if outfilename is not None:
            return outfilename

    return None


def find_module_cpppath(modulename, opt_in=False):
    filepath = _find_module_cpppath(modulename, opt_in)
    if filepath is None:
        raise ImportError(
            "Couldn't find a file matching the module name: "
            + str(modulename)
            + "  (note: opt_in = "
            + str(opt_in)
            + ")"
        )
    return filepath



import os
import sys
import argparse

import cppimport

footer = """
<%
import pybind11
cfg['compiler_args'] = ['-std=c++11']
cfg['include_dirs'] = [pybind11.get_include(), pybind11.get_include(True)]
%>
#include <pybind11/pybind11.h>
PYBIND11_PLUGIN(${fullname}) {
    pybind11::module m("${fullname}", "");
    m.def("main", main);
    return m.ptr();
}
"""


def cpp_run():
    parser = argparse.ArgumentParser(description="Run a C++ file with cppimport")
    parser.add_argument("filename", help="The file to run.")
    parser.add_argument(
        "--add_main_caller",
        "-m",
        action="store_true",
        help="Add a pybind11 function that will call your main()",
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Tell me everything!"
    )
    args = parser.parse_args()

    if args.verbose:
        cppimport.set_quiet(False)

    filename = args.filename
    filedir = os.path.dirname(filename)
    filebasename = os.path.basename(filename)
    module_name, file_extension = os.path.splitext(filebasename)

    if args.add_main_caller:
        cpprun_dir = ".cpprunfiles"
        if not os.path.exists(cpprun_dir):
            os.makedirs(cpprun_dir)
        src = os.path.join(cpprun_dir, filebasename)
        open(src, "w").write(open(filename, "r").read() + footer)
        sys.path.append(cpprun_dir)
    else:
        sys.path.append(filedir)

    module = cppimport.imp(module_name)

    if args.verbose:
        print("Launching!")
    module.main()


if __name__ == "__main__":
    cpp_run()



from cppimport.config import (
    set_quiet,
    force_rebuild,
    file_exts,
    turn_off_strict_prototypes,
    set_rtld_flags,
)
from cppimport.importer import imp, imp_from_filepath, build
from cppimport.templating import setup_pybind11
from cppimport.importer import imp as cppimport



import os
import sys


def get_rendered_source_filepath(filepath):
    dirname = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    return os.path.join(dirname, ".rendered." + filename)


class BuildArgs(dict):
    _key_mapping = {
        "compiler_args": "extra_compile_args",
        "linker_args": "extra_link_args",
    }

    def __getitem__(self, key):
        return super(BuildArgs, self).__getitem__(self._key_mapping.get(key, key))

    def __setitem__(self, key, value):
        super(BuildArgs, self).__setitem__(self._key_mapping.get(key, key), value)


def setup_pybind11(cfg):
    import pybind11

    cfg["include_dirs"] += [pybind11.get_include(), pybind11.get_include(True)]
    # Prefix with c++11 arg instead of suffix so that if a user specifies c++14
    # (or later!) then it won't be overridden.
    cfg["compiler_args"] = ["-std=c++11", "-fvisibility=hidden"] + cfg["compiler_args"]


def run_templating(module_data):
    import mako.template
    import mako.runtime
    import mako.exceptions
    import mako.lookup

    if sys.version_info[0] == 2:
        import StringIO as io
    else:
        import io

    module_data["cfg"] = BuildArgs(
        sources=[],
        include_dirs=[],
        extra_compile_args=[],
        libraries=[],
        library_dirs=[],
        extra_link_args=[],
        dependencies=[],
        parallel=False,
    )
    module_data["setup_pybind11"] = setup_pybind11
    buf = io.StringIO()
    ctx = mako.runtime.Context(buf, **module_data)

    filepath = module_data["filepath"]
    lookup = mako.lookup.TemplateLookup(directories=[os.path.dirname(filepath)])
    tmpl = mako.template.Template(filename=filepath, lookup=lookup)

    try:
        tmpl.render_context(ctx)
    except Exception:
        print(mako.exceptions.text_error_template().render())

    rendered_src_filepath = get_rendered_source_filepath(filepath)

    with open(rendered_src_filepath, "w", newline="") as f:
        f.write(buf.getvalue())

    module_data["rendered_src_filepath"] = rendered_src_filepath



import os
import io
import sys
import copy
import subprocess
import contextlib

import cppimport
import cppimport.build_module
import cppimport.templating
import cppimport.import_hook

cppimport.set_quiet(False)


@contextlib.contextmanager
def appended(filename, text):
    with open(filename, "r") as f:
        orig = f.read()
    with open(filename, "a") as f:
        f.write(text)
    try:
        yield
    finally:
        with open(filename, "w") as f:
            f.write(orig)


def subprocess_check(test_code, returncode=0):
    p = subprocess.run(
        ["python", "-c", test_code], cwd=os.path.dirname(__file__), capture_output=True
    )
    print(p.stdout.decode("utf-8"))
    print(p.stderr.decode("utf-8"))
    assert p.returncode == returncode


def test_redirected_stream():
    sys.stderr = io.StringIO()
    with cppimport.build_module.stdchannel_redirected("stdout") as s:
        with cppimport.build_module.stdchannel_redirected("stderr"):
            print("EEEP!")
    assert s.getvalue() == "EEEP!\n"


def test_find_module_cpppath():
    mymodule_loc = cppimport.find.find_module_cpppath("mymodule")
    mymodule_dir = os.path.dirname(mymodule_loc)
    assert os.path.basename(mymodule_loc) == "mymodule.cpp"

    apackage = cppimport.find.find_module_cpppath("apackage.mymodule")
    apackage_correct = os.path.join(mymodule_dir, "apackage", "mymodule.cpp")
    assert apackage == apackage_correct

    inner = cppimport.find.find_module_cpppath("apackage.inner.mymodule")
    inner_correct = os.path.join(mymodule_dir, "apackage", "inner", "mymodule.cpp")
    assert inner == inner_correct


def test_get_rendered_source_filepath():
    rendered_path = cppimport.templating.get_rendered_source_filepath("abc.cpp")
    assert rendered_path == ".rendered.abc.cpp"


def module_tester(mod, cheer=False):
    assert mod.add(1, 2) == 3
    if cheer:
        mod.Thing().cheer()


def test_mymodule():
    mymodule = cppimport.imp("mymodule")
    module_tester(mymodule)


def test_package_mymodule():
    mymodule = cppimport.imp("apackage.mymodule")
    module_tester(mymodule)


def test_inner_package_mymodule():
    mymodule = cppimport.imp("apackage.inner.mymodule")
    module_tester(mymodule)


def test_with_file_in_syspath():
    orig_sys_path = copy.copy(sys.path)
    sys.path.append(os.path.join(os.path.dirname(__file__), "mymodule.cpp"))
    cppimport.imp("mymodule")
    sys.path = orig_sys_path


def test_rebuild_after_failed_compile():
    cppimport.imp("mymodule")
    test_code = """
import cppimport; mymodule = cppimport.imp("mymodule");assert(mymodule.add(1,2) == 3)
"""
    with appended("tests/mymodule.cpp", ";asdf;"):
        subprocess_check(test_code, 1)
    subprocess_check(test_code, 0)


add_to_thing = """
#include <iostream>
struct Thing {
    void cheer() {
        std::cout << "WAHHOOOO" << std::endl;
    }
};
#define THING_DEFINED
"""


def test_no_rebuild_if_no_deps_change():
    cppimport.imp("mymodule")
    test_code = """
import cppimport;
mymodule = cppimport.imp("mymodule");
assert(not hasattr(mymodule, 'Thing'))
"""
    with appended("tests/thing2.h", add_to_thing):
        subprocess_check(test_code)


def test_rebuild_header_after_change():
    cppimport.imp("mymodule")
    test_code = """
import cppimport;
cppimport.set_quiet(False);
mymodule = cppimport.imp("mymodule");
mymodule.Thing().cheer()
"""
    with appended("tests/thing.h", add_to_thing):
        subprocess_check(test_code)
    assert open("tests/thing.h", "r").read() == ""


def test_raw_extensions():
    raw_extension = cppimport.imp("raw_extension")
    assert raw_extension.add(1, 2) == 3


def test_extra_sources():
    mod = cppimport.imp("extra_sources")
    assert mod.square_sum(3, 4) == 25


# TODO: cpprun is incomplete and possibly not a good idea...
# def test_cpprun():
#     p = subprocess.Popen([
#         'cpprun', '-m', 'free_module.cpp'
#     ], cwd = os.path.dirname(__file__), stdout = subprocess.PIPE)
#     p.wait()
#     assert(b'HI!\n' == p.stdout.read())


def test_import_hook():
    # Force rebuild to make sure we're not just reloading the already compiled
    # module from disk
    cppimport.force_rebuild(True)
    import hook_test

    cppimport.force_rebuild(False)
    hook_test



from setuptools import setup

version = open("VERSION").read()

description = open("README.md").read()

setup(
    packages=["cppimport"],
    install_requires=["mako", "pybind11"],
    zip_safe=False,
    name="cppimport",
    version=version,
    description="Import C++ files directly from Python!",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/tbenthompson/cppimport",
    author="T. Ben Thompson",
    author_email="t.ben.thompson@gmail.com",
    license="MIT",
    platforms=["any"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Topic :: Software Development",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
    ],
)



import os
import sys
import sysconfig
import importlib

import cppimport.config
import cppimport.checksum
import cppimport.find

quiet_print = cppimport.config.quiet_print


def get_module_name(full_module_name):
    return full_module_name.split(".")[-1]


def get_extension_suffix():
    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
    if ext_suffix is None:
        ext_suffix = sysconfig.get_config_var("SO")
    return ext_suffix


def setup_module_data(fullname, filepath):
    module_data = dict()
    module_data["fullname"] = fullname
    module_data["filepath"] = filepath
    module_data["filedirname"] = os.path.dirname(module_data["filepath"])
    module_data["filebasename"] = os.path.basename(module_data["filepath"])
    module_data["ext_name"] = get_module_name(fullname) + get_extension_suffix()
    module_data["ext_path"] = os.path.join(
        os.path.dirname(filepath), module_data["ext_name"]
    )
    return module_data


def _load_module(module_data):
    module_data["module"] = importlib.import_module(module_data["fullname"])


def load_module(module_data):
    if hasattr(sys, "getdlopenflags"):
        old_flags = sys.getdlopenflags()
        new_flags = old_flags | cppimport.config.rtld_flags
        sys.setdlopenflags(new_flags)
        _load_module(module_data)
        sys.setdlopenflags(old_flags)
    else:
        _load_module(module_data)


def check_checksum(module_data):
    if cppimport.config.should_force_rebuild:
        return False
    if not cppimport.checksum.is_checksum_current(module_data):
        return False
    quiet_print(
        "Matching checksum for " + module_data["filepath"] + " --> not compiling"
    )
    return True


def try_load(module_data):
    try:
        load_module(module_data)
        return True
    except ImportError:
        quiet_print(
            "ImportError during import with matching checksum. Trying to rebuild."
        )
        return False


def template_and_build(filepath, module_data):
    # Don't import until here to reduce startup time.
    import cppimport.templating as templating
    import cppimport.build_module as build_module

    quiet_print("Compiling " + filepath)
    templating.run_templating(module_data)
    build_module.build_module(module_data)
    if cppimport.config.quiet:
        os.remove(module_data["rendered_src_filepath"])
    cppimport.checksum.checksum_save(module_data)


def imp_from_filepath(filepath, fullname=None):
    if fullname is None:
        fullname = os.path.splitext(os.path.basename(filepath))[0]
    module_data = setup_module_data(fullname, filepath)
    if not check_checksum(module_data) or not try_load(module_data):
        template_and_build(filepath, module_data)
        load_module(module_data)
    return module_data["module"]


def imp(fullname, opt_in=False):
    # Search through sys.path to find a file that matches the module
    filepath = cppimport.find.find_module_cpppath(fullname, opt_in)
    return imp_from_filepath(filepath, fullname)


def build(fullname):
    # Search through sys.path to find a file that matches the module
    filepath = cppimport.find.find_module_cpppath(fullname)
    module_data = setup_module_data(fullname, filepath)
    if not check_checksum(module_data):
        template_and_build(filepath, module_data)

    # Return the path to the built module
    return module_data["ext_path"]



import json
import hashlib
import struct

import cppimport.find
import cppimport.config
from cppimport.filepaths import make_absolute


_TAG = b"cppimport"
_FMT = struct.Struct("q" + str(len(_TAG)) + "s")


def calc_cur_checksum(file_lst, module_data):
    text = b""
    for filepath in file_lst:
        try:
            with open(filepath, "rb") as f:
                text += f.read()
        except OSError as e:
            cppimport.config.quiet_print(
                "Checksummed file not found while checking cppimport checksum "
                "(%s); rebuilding." % e
            )
            return None
    return hashlib.md5(text).hexdigest()


def load_checksum_trailer(module_data):
    try:
        with open(module_data["ext_path"], "rb") as f:
            f.seek(-_FMT.size, 2)
            json_len, tag = _FMT.unpack(f.read(_FMT.size))
            if tag != _TAG:
                cppimport.config.quiet_print("Missing trailer tag")
                return None, None
            f.seek(-(_FMT.size + json_len), 2)
            json_s = f.read(json_len)
    except FileNotFoundError:
        cppimport.config.quiet_print("Failed to find compiled extension; rebuilding.")
        return None, None

    try:
        deps, old_checksum = json.loads(json_s)
    except ValueError:
        cppimport.config.quiet_print(
            "Failed to load checksum trailer info from already existing "
            "compiled extension; rebuilding."
        )
        return None, None
    return deps, old_checksum


# Use a checksum to see if the file has been changed since the last compilation
def is_checksum_current(module_data):
    deps, old_checksum = load_checksum_trailer(module_data)
    if old_checksum is None:
        return False  # Already logged error in load_checksum_trailer.
    return old_checksum == calc_cur_checksum(deps, module_data)


def save_checksum_trailer(module_data, dep_filepaths, cur_checksum):
    # We can just append the checksum to the shared object; this is effectively
    # legal (see e.g. https://stackoverflow.com/questions/10106447).
    dump = json.dumps([dep_filepaths, cur_checksum]).encode("ascii")
    dump += _FMT.pack(len(dump), _TAG)
    with open(module_data["ext_path"], "ab") as file:
        file.write(dump)


def checksum_save(module_data):
    dep_filepaths = (
        [
            make_absolute(module_data["filedirname"], d)
            for d in module_data["cfg"].get("dependencies", [])
        ]
        + module_data["extra_source_filepaths"]
        + [module_data["filepath"]]
    )
    cur_checksum = calc_cur_checksum(dep_filepaths, module_data)
    save_checksum_trailer(module_data, dep_filepaths, cur_checksum)
